package solutions.bitbadger.documents

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertEquals
import kotlin.test.assertTrue

class QueryTest {

    /** Test table name */
    private val tbl = "test_table"

    /** Dummy connection string for PostgreSQL */
    private val pg = ":postgresql:"

    /** Dummy connection string for SQLite */
    private val lite = ":sqlite:"

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        Configuration.connectionString = null
    }

    @Test
    @DisplayName("statementWhere generates correctly")
    fun statementWhere() =
        assertEquals("x WHERE y", Query.statementWhere("x", "y"), "Statements not combined correctly")

    // ~~~ Where ~~~

    @Test
    @DisplayName("Where.byFields is blank when given no fields")
    fun whereByFieldsBlankIfEmpty() =
        assertEquals("", Query.Where.byFields(listOf()))

    @Test
    @DisplayName("Where.byFields generates one numeric field (PostgreSQL)")
    fun whereByFieldsOneFieldPostgres() {
        Configuration.connectionString = pg
        assertEquals("(data->>'it')::numeric = :that", Query.Where.byFields(listOf(Field.equal("it", 9, ":that"))))
    }

    @Test
    @DisplayName("Where.byFields generates one alphanumeric field (PostgreSQL)")
    fun whereByFieldsOneAlphaFieldPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'it' = :that", Query.Where.byFields(listOf(Field.equal("it", "", ":that"))))
    }

    @Test
    @DisplayName("Where.byFields generates one field (SQLite)")
    fun whereByFieldsOneFieldSQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'it' = :that", Query.Where.byFields(listOf(Field.equal("it", "", ":that"))))
    }

    @Test
    @DisplayName("Where.byFields generates multiple fields w/ default match (PostgreSQL)")
    fun whereByFieldsMultipleDefaultPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'1' = :one AND (data->>'2')::numeric = :two AND data->>'3' = :three",
            Query.Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three"))))
    }

    @Test
    @DisplayName("Where.byFields generates multiple fields w/ default match (SQLite)")
    fun whereByFieldsMultipleDefaultSQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'1' = :one AND data->>'2' = :two AND data->>'3' = :three",
            Query.Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three"))))
    }

    @Test
    @DisplayName("Where.byFields generates multiple fields w/ ANY match (PostgreSQL)")
    fun whereByFieldsMultipleAnyPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'1' = :one OR (data->>'2')::numeric = :two OR data->>'3' = :three",
            Query.Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")),
                FieldMatch.ANY))
    }

    @Test
    @DisplayName("Where.byFields generates multiple fields w/ ANY match (SQLite)")
    fun whereByFieldsMultipleAnySQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'1' = :one OR data->>'2' = :two OR data->>'3' = :three",
            Query.Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")),
                FieldMatch.ANY))
    }

    @Test
    @DisplayName("Where.byId generates defaults for alphanumeric key (PostgreSQL)")
    fun whereByIdDefaultAlphaPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'id' = :id", Query.Where.byId(docId = ""))
    }

    @Test
    @DisplayName("Where.byId generates defaults for numeric key (PostgreSQL)")
    fun whereByIdDefaultNumericPostgres() {
        Configuration.connectionString = pg
        assertEquals("(data->>'id')::numeric = :id", Query.Where.byId(docId = 5))
    }

    @Test
    @DisplayName("Where.byId generates defaults (SQLite)")
    fun whereByIdDefaultSQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'id' = :id", Query.Where.byId(docId = ""))
    }

    @Test
    @DisplayName("Where.byId generates named ID (PostgreSQL)")
    fun whereByIdDefaultNamedPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'id' = :key", Query.Where.byId<String>(":key"))
    }

    @Test
    @DisplayName("Where.byId generates named ID (SQLite)")
    fun whereByIdDefaultNamedSQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'id' = :key", Query.Where.byId<String>(":key"))
    }

    @Test
    @DisplayName("Where.jsonContains generates defaults (PostgreSQL)")
    fun whereJsonContainsDefaultPostgres() {
        Configuration.connectionString = pg
        assertEquals("data @> :criteria", Query.Where.jsonContains())
    }

    @Test
    @DisplayName("Where.jsonContains generates named parameter (PostgreSQL)")
    fun whereJsonContainsNamedPostgres() {
        Configuration.connectionString = pg
        assertEquals("data @> :it", Query.Where.jsonContains(":it"))
    }

    @Test
    @DisplayName("Where.jsonContains fails (SQLite)")
    fun whereJsonContainsFailsSQLite() {
        Configuration.connectionString = lite
        assertThrows<DocumentException> { Query.Where.jsonContains() }
    }

    @Test
    @DisplayName("Where.jsonPathMatch generates defaults (PostgreSQL)")
    fun whereJsonPathMatchDefaultPostgres() {
        Configuration.connectionString = pg
        assertEquals("jsonb_path_exists(data, :path::jsonpath)", Query.Where.jsonPathMatches())
    }

    @Test
    @DisplayName("Where.jsonPathMatch generates named parameter (PostgreSQL)")
    fun whereJsonPathMatchNamedPostgres() {
        Configuration.connectionString = pg
        assertEquals("jsonb_path_exists(data, :jp::jsonpath)", Query.Where.jsonPathMatches(":jp"))
    }

    @Test
    @DisplayName("Where.jsonPathMatch fails (SQLite)")
    fun whereJsonPathFailsSQLite() {
        Configuration.connectionString = lite
        assertThrows<DocumentException> { Query.Where.jsonPathMatches() }
    }

    // ~~~ root functions ~~~

    @Test
    @DisplayName("byId generates a numeric ID query (PostgreSQL)")
    fun byIdNumericPostgres() {
        Configuration.connectionString = pg
        assertEquals("test WHERE (data->>'id')::numeric = :id", Query.byId("test", 9))
    }

    @Test
    @DisplayName("byId generates an alphanumeric ID query (PostgreSQL)")
    fun byIdAlphaPostgres() {
        Configuration.connectionString = pg
        assertEquals("unit WHERE data->>'id' = :id", Query.byId("unit", "18"))
    }

    @Test
    @DisplayName("byId generates ID query (SQLite)")
    fun byIdSQLite() {
        Configuration.connectionString = lite
        assertEquals("yo WHERE data->>'id' = :id", Query.byId("yo", 27))
    }

    @Test
    @DisplayName("byFields generates default field query (PostgreSQL)")
    fun byFieldsMultipleDefaultPostgres() {
        Configuration.connectionString = pg
        assertEquals("this WHERE data->>'a' = :the_a AND (data->>'b')::numeric = :b_value",
            Query.byFields("this", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value"))))
    }

    @Test
    @DisplayName("byFields generates default field query (SQLite)")
    fun byFieldsMultipleDefaultSQLite() {
        Configuration.connectionString = lite
        assertEquals("this WHERE data->>'a' = :the_a AND data->>'b' = :b_value",
            Query.byFields("this", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value"))))
    }

    @Test
    @DisplayName("byFields generates ANY field query (PostgreSQL)")
    fun byFieldsMultipleAnyPostgres() {
        Configuration.connectionString = pg
        assertEquals("that WHERE data->>'a' = :the_a OR (data->>'b')::numeric = :b_value",
            Query.byFields("that", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")),
                FieldMatch.ANY))
    }

    @Test
    @DisplayName("byFields generates ANY field query (SQLite)")
    fun byFieldsMultipleAnySQLite() {
        Configuration.connectionString = lite
        assertEquals("that WHERE data->>'a' = :the_a OR data->>'b' = :b_value",
            Query.byFields("that", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")),
                FieldMatch.ANY))
    }

    // ~~~ Definition ~~~

    @Test
    @DisplayName("Definition.ensureTableFor generates correctly")
    fun ensureTableFor() =
        assertEquals("CREATE TABLE IF NOT EXISTS my.table (data JSONB NOT NULL)",
            Query.Definition.ensureTableFor("my.table", "JSONB"), "CREATE TABLE statement not constructed correctly")

    @Test
    @DisplayName("Definition.ensureTable generates correctly (PostgreSQL)")
    fun ensureTablePostgres() {
        Configuration.connectionString = pg
        assertEquals("CREATE TABLE IF NOT EXISTS $tbl (data JSONB NOT NULL)", Query.Definition.ensureTable(tbl))
    }

    @Test
    @DisplayName("Definition.ensureTable generates correctly (SQLite)")
    fun ensureTableSQLite() {
        Configuration.connectionString = lite
        assertEquals("CREATE TABLE IF NOT EXISTS $tbl (data TEXT NOT NULL)", Query.Definition.ensureTable(tbl))
    }

    @Test
    @DisplayName("Definition.ensureTable fails when no dialect is set")
    fun ensureTableFailsUnknown() {
        assertThrows<DocumentException> { Query.Definition.ensureTable(tbl) }
    }

    @Test
    @DisplayName("Definition.ensureKey generates correctly with schema")
    fun ensureKeyWithSchema() =
        assertEquals("CREATE UNIQUE INDEX IF NOT EXISTS idx_table_key ON test.table ((data->>'id'))",
            Query.Definition.ensureKey("test.table", Dialect.POSTGRESQL),
            "CREATE INDEX for key statement with schema not constructed correctly")

    @Test
    @DisplayName("Definition.ensureKey generates correctly without schema")
    fun ensureKeyWithoutSchema() =
        assertEquals("CREATE UNIQUE INDEX IF NOT EXISTS idx_${tbl}_key ON $tbl ((data->>'id'))",
            Query.Definition.ensureKey(tbl, Dialect.SQLITE),
            "CREATE INDEX for key statement without schema not constructed correctly")

    @Test
    @DisplayName("Definition.ensureIndexOn generates multiple fields and directions")
    fun ensureIndexOnMultipleFields() =
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_table_gibberish ON test.table ((data->>'taco'), (data->>'guac') DESC, (data->>'salsa') ASC)",
            Query.Definition.ensureIndexOn("test.table", "gibberish", listOf("taco", "guac DESC", "salsa ASC"),
                Dialect.POSTGRESQL),
            "CREATE INDEX for multiple field statement not constructed correctly")

    @Test
    @DisplayName("Definition.ensureIndexOn generates nested PostgreSQL field")
    fun ensureIndexOnNestedPostgres() =
        assertEquals("CREATE INDEX IF NOT EXISTS idx_${tbl}_nest ON $tbl ((data#>>'{a,b,c}'))",
            Query.Definition.ensureIndexOn(tbl, "nest", listOf("a.b.c"), Dialect.POSTGRESQL),
            "CREATE INDEX for nested PostgreSQL field incorrect")

    @Test
    @DisplayName("Definition.ensureIndexOn generates nested SQLite field")
    fun ensureIndexOnNestedSQLite() =
        assertEquals("CREATE INDEX IF NOT EXISTS idx_${tbl}_nest ON $tbl ((data->'a'->'b'->>'c'))",
            Query.Definition.ensureIndexOn(tbl, "nest", listOf("a.b.c"), Dialect.SQLITE),
            "CREATE INDEX for nested SQLite field incorrect")

    // ~~~ root functions ~~~

    @Test
    @DisplayName("insert generates no auto ID (PostgreSQL)")
    fun insertNoAutoPostgres() {
        Configuration.connectionString = pg
        assertEquals("INSERT INTO $tbl VALUES (:data)", Query.insert(tbl))
    }

    @Test
    @DisplayName("insert generates no auto ID (SQLite)")
    fun insertNoAutoSQLite() {
        Configuration.connectionString = lite
        assertEquals("INSERT INTO $tbl VALUES (:data)", Query.insert(tbl))
    }

    @Test
    @DisplayName("insert generates auto number (PostgreSQL)")
    fun insertAutoNumberPostgres() {
        Configuration.connectionString = pg
        assertEquals(
            "INSERT INTO $tbl VALUES (:data::jsonb || ('{\"id\":' " +
                    "|| (SELECT COALESCE(MAX((data->>'id')::numeric), 0) + 1 FROM $tbl) || '}')::jsonb)",
            Query.insert(tbl, AutoId.NUMBER))
    }

    @Test
    @DisplayName("insert generates auto number (SQLite)")
    fun insertAutoNumberSQLite() {
        Configuration.connectionString = lite
        assertEquals(
            "INSERT INTO $tbl VALUES (json_set(:data, '$.id', " +
                    "(SELECT coalesce(max(data->>'id'), 0) + 1 FROM $tbl)))",
            Query.insert(tbl, AutoId.NUMBER))
    }

    @Test
    @DisplayName("insert generates auto UUID (PostgreSQL)")
    fun insertAutoUUIDPostgres() {
        Configuration.connectionString = pg
        val query = Query.insert(tbl, AutoId.UUID)
        assertTrue(query.startsWith("INSERT INTO $tbl VALUES (:data::jsonb || '{\"id\":\""),
            "Query start not correct (actual: $query)")
        assertTrue(query.endsWith("\"}')"), "Query end not correct")
    }

    @Test
    @DisplayName("insert generates auto UUID (SQLite)")
    fun insertAutoUUIDSQLite() {
        Configuration.connectionString = lite
        val query = Query.insert(tbl, AutoId.UUID)
        assertTrue(query.startsWith("INSERT INTO $tbl VALUES (json_set(:data, '$.id', '"),
            "Query start not correct (actual: $query)")
        assertTrue(query.endsWith("'))"), "Query end not correct")
    }

    @Test
    @DisplayName("insert generates auto random string (PostgreSQL)")
    fun insertAutoRandomPostgres() {
        try {
            Configuration.connectionString = pg
            Configuration.idStringLength   = 8
            val query = Query.insert(tbl, AutoId.RANDOM_STRING)
            assertTrue(query.startsWith("INSERT INTO $tbl VALUES (:data::jsonb || '{\"id\":\""),
                "Query start not correct (actual: $query)")
            assertTrue(query.endsWith("\"}')"), "Query end not correct")
            assertEquals(8,
                query.replace("INSERT INTO $tbl VALUES (:data::jsonb || '{\"id\":\"", "").replace("\"}')", "").length,
                "Random string length incorrect")
        } finally {
            Configuration.idStringLength = 16
        }
    }

    @Test
    @DisplayName("insert generates auto random string (SQLite)")
    fun insertAutoRandomSQLite() {
        Configuration.connectionString = lite
        val query = Query.insert(tbl, AutoId.RANDOM_STRING)
        assertTrue(query.startsWith("INSERT INTO $tbl VALUES (json_set(:data, '$.id', '"),
            "Query start not correct (actual: $query)")
        assertTrue(query.endsWith("'))"), "Query end not correct")
        assertEquals(Configuration.idStringLength,
            query.replace("INSERT INTO $tbl VALUES (json_set(:data, '$.id', '", "").replace("'))", "").length,
            "Random string length incorrect")
    }

    @Test
    @DisplayName("insert fails when no dialect is set")
    fun insertFailsUnknown() {
        assertThrows<DocumentException> { Query.insert(tbl) }
    }

    @Test
    @DisplayName("save generates correctly")
    fun save() {
        Configuration.connectionString = pg
        assertEquals(
            "INSERT INTO $tbl VALUES (:data) ON CONFLICT ((data->>'id')) DO UPDATE SET data = EXCLUDED.data",
            Query.save(tbl), "INSERT ON CONFLICT UPDATE statement not constructed correctly")
    }

    @Test
    @DisplayName("count generates correctly")
    fun count() =
        assertEquals("SELECT COUNT(*) AS it FROM $tbl", Query.count(tbl), "Count query not constructed correctly")

    @Test
    @DisplayName("exists generates correctly")
    fun exists() =
        assertEquals("SELECT EXISTS (SELECT 1 FROM $tbl WHERE turkey) AS it", Query.exists(tbl, "turkey"),
            "Exists query not constructed correctly")

    @Test
    @DisplayName("find generates correctly")
    fun find() =
        assertEquals("SELECT data FROM $tbl", Query.find(tbl), "Find query not constructed correctly")

    @Test
    @DisplayName("update generates successfully")
    fun update() =
        assertEquals("UPDATE $tbl SET data = :data", Query.update(tbl), "Update query not constructed correctly")

    @Test
    @DisplayName("delete generates successfully")
    fun delete() =
        assertEquals("DELETE FROM $tbl", Query.delete(tbl), "Delete query not constructed correctly")

    @Test
    @DisplayName("orderBy generates for no fields")
    fun orderByNone() {
        assertEquals("", Query.orderBy(listOf(), Dialect.POSTGRESQL), "ORDER BY should have been blank (PostgreSQL)")
        assertEquals("", Query.orderBy(listOf(), Dialect.SQLITE), "ORDER BY should have been blank (SQLite)")
    }

    @Test
    @DisplayName("orderBy generates single, no direction for PostgreSQL")
    fun orderBySinglePostgres() =
        assertEquals(" ORDER BY data->>'TestField'",
            Query.orderBy(listOf(Field.named("TestField")), Dialect.POSTGRESQL), "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates single, no direction for SQLite")
    fun orderBySingleSQLite() =
        assertEquals(" ORDER BY data->>'TestField'", Query.orderBy(listOf(Field.named("TestField")), Dialect.SQLITE),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates multiple with direction for PostgreSQL")
    fun orderByMultiplePostgres() =
        assertEquals(" ORDER BY data#>>'{Nested,Test,Field}' DESC, data->>'AnotherField', data->>'It' DESC",
            Query.orderBy(
                listOf(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                Dialect.POSTGRESQL),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates multiple with direction for SQLite")
    fun orderByMultipleSQLite() =
        assertEquals(" ORDER BY data->'Nested'->'Test'->>'Field' DESC, data->>'AnotherField', data->>'It' DESC",
            Query.orderBy(
                listOf(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                Dialect.SQLITE),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates numeric ordering PostgreSQL")
    fun orderByNumericPostgres() =
        assertEquals(" ORDER BY (data->>'Test')::numeric",
            Query.orderBy(listOf(Field.named("n:Test")), Dialect.POSTGRESQL), "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates numeric ordering for SQLite")
    fun orderByNumericSQLite() =
        assertEquals(" ORDER BY data->>'Test'", Query.orderBy(listOf(Field.named("n:Test")), Dialect.SQLITE),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates case-insensitive ordering for PostgreSQL")
    fun orderByCIPostgres() =
        assertEquals(" ORDER BY LOWER(data#>>'{Test,Field}') DESC NULLS FIRST",
            Query.orderBy(listOf(Field.named("i:Test.Field DESC NULLS FIRST")), Dialect.POSTGRESQL),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates case-insensitive ordering for SQLite")
    fun orderByCISQLite() =
        assertEquals(" ORDER BY data->'Test'->>'Field' COLLATE NOCASE ASC NULLS LAST",
            Query.orderBy(listOf(Field.named("i:Test.Field ASC NULLS LAST")), Dialect.SQLITE),
            "ORDER BY not constructed correctly")
}
